#  tests for pymc3-3.5-py27_0 (this is a generated file);
print('===== testing package: pymc3-3.5-py27_0 =====');
print('running run_test.py');
#  --- run_test.py (begin) ---
import os
# theano >= 1.0.0
os.environ['MKL_THREADING_LAYER']='GNU'

import pymc3
import pymc3.backends
import pymc3.distributions
import pymc3.examples
import pymc3.external
import pymc3.glm
import pymc3.gp
import pymc3.plots
import pymc3.step_methods
import pymc3.step_methods.hmc
import pymc3.tests
import pymc3.tuning
import pymc3.variational
#  --- run_test.py (end) ---

print('===== pymc3-3.5-py27_0 OK =====');
print("import: 'pymc3'")
import pymc3

print("import: 'pymc3.backends'")
import pymc3.backends

print("import: 'pymc3.distributions'")
import pymc3.distributions

print("import: 'pymc3.examples'")
import pymc3.examples

print("import: 'pymc3.external'")
import pymc3.external

print("import: 'pymc3.glm'")
import pymc3.glm

print("import: 'pymc3.gp'")
import pymc3.gp

print("import: 'pymc3.plots'")
import pymc3.plots

print("import: 'pymc3.step_methods'")
import pymc3.step_methods

print("import: 'pymc3.step_methods.hmc'")
import pymc3.step_methods.hmc

print("import: 'pymc3.tests'")
import pymc3.tests

print("import: 'pymc3.tuning'")
import pymc3.tuning

print("import: 'pymc3.variational'")
import pymc3.variational

